# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2022 Alban 'spl0k' Féron
# Copyright (C) 2024 Iván Ávalos
#
# Distributed under terms of the GNU AGPLv3 license.

import hashlib
import logging
import requests
import json
from urllib.parse import urljoin

logger = logging.getLogger(__name__)


class ListenBrainz:
    def __init__(self, config, user):
        if config["api_url"] is not None:
            self.__api_url = config["api_url"]
            self.__enabled = True
        else:
            self.__enabled = False
        self.__user = user

    def link_account(self, token):
        if not self.__enabled:
            return False, "No ListenBrainz URL set"

        res = self.__api_request(False, "/1/validate-token", token)
        if not res:
            return False, "Error connecting to ListenBrainz"
        else:
            if "valid" in res and res["valid"]:
                self.__user.listenbrainz_session = token
                self.__user.listenbrainz_status = True
                self.__user.save()
                return True, "OK"
            else:
                return False, f"Error: {res['message']}"

    def unlink_account(self):
        self.__user.listenbrainz_session = None
        self.__user.listenbrainz_status = True
        self.__user.save()

    def now_playing(self, track):
        if not self.__enabled:
            return

        self.__api_request(
            True,
            "/1/submit-listens",
            self.__user.listenbrainz_session,
            listen_type="playing_now",
            payload=[
                {
                    "track_metadata": {
                        "artist_name": track.album.artist.name,
                        "track_name": track.title,
                        "release_name": track.album.name,
                        "additional_info": {
                            "media_player": "Supysonic",
                            "duration_ms": track.duration,
                        },
                    },
                }
            ],
        )

    def scrobble(self, track, ts):
        if not self.__enabled:
            return

        self.__api_request(
            True,
            "/1/submit-listens",
            self.__user.listenbrainz_session,
            listen_type="single",
            payload=[
                {
                    "listened_at": ts,
                    "track_metadata": {
                        "artist_name": track.album.artist.name,
                        "track_name": track.title,
                        "release_name": track.album.name,
                        "additional_info": {
                            "media_player": "Supysonic",
                            "duration_ms": track.duration,
                        },
                    },
                }
            ],
        )

    def __api_request(self, write, route, token, **kwargs):
        if not self.__enabled or not token:
            return

        headers = {"Content-Type": "application/json"}
        headers["Authorization"] = "Token {0}".format(token)

        try:
            if write:
                r = requests.post(
                    urljoin(self.__api_url, route),
                    headers=headers,
                    data=json.dumps(kwargs),
                    timeout=5,
                )
            else:
                r = requests.get(
                    urljoin(self.__api_url, route),
                    headers=headers,
                    data=json.dumps(kwargs),
                    timeout=5,
                )

            r.raise_for_status()
        except requests.HTTPError as e:
            status_code = e.response.status_code
            if status_code == 401:  # Unauthorized
                self.__user.listenbrainz_status = False
                self.__user.save()
            message = e.response.json().get("error", "")
            logger.warning("ListenBrainz error %i: %s", status_code, message)
            return None
        except requests.exceptions.RequestException as e:
            logger.warning("Error while connecting to ListenBrainz: " + str(e))
            return None

        return r.json()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2017-2024 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import logging

from ..testbase import TestBase


class FrontendTestBase(TestBase):
    __with_webui__ = True

    def setUp(self):
        super().setUp()
        logging.getLogger("supysonic.frontend.user").addHandler(logging.NullHandler())
        self._patch_client()

    def _login(self, username, password):
        return self.client.post(
            "/user/login",
            data={"user": username, "password": password},
            follow_redirects=True,
        )

    def _logout(self):
        return self.client.get("/user/logout", follow_redirects=True)



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2023 Alban 'spl0k' Féron
#               2018-2019 Carey 'pR0Ps' Metcalfe
#                    2017 Óscar García Amor
#
# Distributed under terms of the GNU AGPLv3 license.

import logging
import mimetypes

from flask import Flask
from logging.handlers import TimedRotatingFileHandler
from os import makedirs, path

from .config import IniConfig
from .cache import Cache
from .db import init_database, open_connection, close_connection
from .utils import get_secret_key

logger = logging.getLogger(__package__)


def create_application(config=None):
    global app

    # Flask!
    app = Flask(__name__)
    app.config.from_object("supysonic.config.DefaultConfig")

    if not config:  # pragma: nocover
        config = IniConfig.from_common_locations()
    app.config.from_object(config)

    # Set loglevel
    logfile = app.config["WEBAPP"]["log_file"]
    if logfile:  # pragma: nocover
        if app.config["WEBAPP"]["log_rotate"]:
            handler = TimedRotatingFileHandler(logfile, when="midnight")
        else:
            handler = logging.FileHandler(logfile)
        handler.setFormatter(
            logging.Formatter("%(asctime)s [%(levelname)s] %(message)s")
        )
        logger.addHandler(handler)
    loglevel = app.config["WEBAPP"]["log_level"]
    if loglevel:
        logger.setLevel(getattr(logging, loglevel.upper(), logging.NOTSET))

    # Initialize database
    init_database(app.config["BASE"]["database_uri"])
    if not app.testing:

        def open_conn():  # Just to discard the return value
            open_connection()

        app.before_request(open_conn)
        app.teardown_request(lambda exc: close_connection())

    # Insert unknown mimetypes
    for k, v in app.config["MIMETYPES"].items():
        extension = "." + k.lower()
        if extension not in mimetypes.types_map:
            mimetypes.add_type(v, extension, False)

    # Initialize Cache objects
    # Max size is MB in the config file but Cache expects bytes
    cache_dir = app.config["WEBAPP"]["cache_dir"]
    max_size_cache = app.config["WEBAPP"]["cache_size"] * 1024**2
    max_size_transcodes = app.config["WEBAPP"]["transcode_cache_size"] * 1024**2
    app.cache = Cache(path.join(cache_dir, "cache"), max_size_cache)
    app.transcode_cache = Cache(path.join(cache_dir, "transcodes"), max_size_transcodes)

    # Test for the cache directory
    cache_path = app.config["WEBAPP"]["cache_dir"]
    if not path.exists(cache_path):
        makedirs(cache_path)  # pragma: nocover

    # Read or create secret key
    app.secret_key = get_secret_key("cookies_secret")

    # Import app sections
    if app.config["WEBAPP"]["mount_webui"]:
        from .frontend import frontend

        app.register_blueprint(frontend)
    if app.config["WEBAPP"]["mount_api"]:
        from .api import api

        app.register_blueprint(api, url_prefix="/rest")

    if not app.testing:
        close_connection()

    return app



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2018 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import unittest
import shutil
import tempfile

from supysonic.db import init_database, release_database
from supysonic.web import create_application

from ..testbase import TestConfig


class SecretTestCase(unittest.TestCase):
    def setUp(self):
        self.__dir = tempfile.mkdtemp()
        self.config = TestConfig(False, False)
        self.config.BASE["database_uri"] = "sqlite://"
        self.config.WEBAPP["cache_dir"] = self.__dir

        init_database(self.config.BASE["database_uri"])
        release_database()

    def tearDown(self):
        shutil.rmtree(self.__dir)

    def test_key(self):
        app1 = create_application(self.config)
        release_database()

        app2 = create_application(self.config)
        release_database()

        self.assertEqual(app1.secret_key, app2.secret_key)


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

from flask import request
from functools import wraps

from ..db import User
from ..managers.user import UserManager

from . import decode_password, api_routing
from .exceptions import Forbidden


def admin_only(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        if not request.user.admin:
            raise Forbidden()
        return f(*args, **kwargs)

    return decorated


@api_routing("/getUser")
def user_info():
    username = request.values["username"]

    if username != request.user.name and not request.user.admin:
        raise Forbidden()

    user = User.get(name=username)
    return request.formatter("user", user.as_subsonic_user())


@api_routing("/getUsers")
@admin_only
def users_info():
    return request.formatter(
        "users", {"user": [u.as_subsonic_user() for u in User.select()]}
    )


def get_roles_dict():
    roles = {}
    for role in ("admin", "jukebox"):
        value = request.values.get(role + "Role")
        value = value in (True, "True", "true", 1, "1")
        roles[role] = value

    return roles


@api_routing("/createUser")
@admin_only
def user_add():
    username = request.values["username"]
    password = request.values["password"]
    email = request.values["email"]
    roles = get_roles_dict()

    password = decode_password(password)
    UserManager.add(username, password, mail=email, **roles)

    return request.formatter.empty


@api_routing("/deleteUser")
@admin_only
def user_del():
    username = request.values["username"]
    UserManager.delete_by_name(username)

    return request.formatter.empty


@api_routing("/changePassword")
def user_changepass():
    username = request.values["username"]
    password = request.values["password"]

    if username != request.user.name and not request.user.admin:
        raise Forbidden()

    password = decode_password(password)
    UserManager.change_password2(username, password)

    return request.formatter.empty


@api_routing("/updateUser")
@admin_only
def user_edit():
    username = request.values["username"]
    user = User.get(name=username)

    if "password" in request.values:
        password = decode_password(request.values["password"])
        UserManager.change_password2(user, password)

    email, admin, jukebox = map(
        request.values.get, ("email", "adminRole", "jukeboxRole")
    )
    if email is not None:
        user.mail = email

    if admin is not None:
        admin = admin in (True, "True", "true", 1, "1")
        user.admin = admin

    if jukebox is not None:
        jukebox = jukebox in (True, "True", "true", 1, "1")
        user.jukebox = jukebox

    user.save()

    return request.formatter.empty



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2021 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import os.path

from shutil import rmtree
from setuptools import setup
from setuptools.command.sdist import sdist as _sdist


class sdist(_sdist):
    def make_release_tree(self, base_dir, files):
        super().make_release_tree(base_dir, files)

        man_dir = os.path.join(base_dir, "man")
        doctrees_dir = os.path.join(man_dir, ".doctrees")
        self.spawn(["sphinx-build", "-q", "-b", "man", "docs", man_dir])
        rmtree(doctrees_dir)


if __name__ == "__main__":
    setup(
        cmdclass={"sdist": sdist},
    )



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import uuid

from flask import Response, flash, redirect, render_template, request, url_for
from functools import wraps

from ..db import Playlist

from . import frontend


@frontend.route("/playlist")
def playlist_index():
    return render_template(
        "playlists.html",
        mine=Playlist.select().where(Playlist.user == request.user),
        others=Playlist.select().where(Playlist.user != request.user, Playlist.public),
    )


def resolve_and_inject_playlist(func):
    @wraps(func)
    def decorated(uid):
        try:
            uid = uuid.UUID(uid)
        except ValueError:
            flash("Invalid playlist id")
            return redirect(url_for("frontend.playlist_index"))

        try:
            playlist = Playlist[uid]
        except Playlist.DoesNotExist:
            flash("Unknown playlist")
            return redirect(url_for("frontend.playlist_index"))

        return func(uid, playlist)

    return decorated


@frontend.route("/playlist/<uid>")
@resolve_and_inject_playlist
def playlist_details(uid, playlist):
    return render_template("playlist.html", playlist=playlist)


@frontend.route("/playlist/<uid>/export")
@resolve_and_inject_playlist
def playlist_export(uid, playlist):
    return Response(
        render_template("playlist_export.m3u", playlist=playlist),
        mimetype="audio/mpegurl",
        headers={"Content-disposition": f"attachment; filename={playlist.name}.m3u"},
    )


@frontend.route("/playlist/<uid>", methods=["POST"])
@resolve_and_inject_playlist
def playlist_update(uid, playlist):
    if playlist.user_id != request.user.id:
        flash("You're not allowed to edit this playlist")
    elif not request.form.get("name"):
        flash("Missing playlist name")
    else:
        playlist.name = request.form.get("name")
        playlist.public = request.form.get("public") in (
            True,
            "True",
            1,
            "1",
            "on",
            "checked",
        )
        playlist.save()
        flash("Playlist updated.")

    return playlist_details(str(uid))


@frontend.route("/playlist/del/<uid>")
@resolve_and_inject_playlist
def playlist_delete(uid, playlist):
    if playlist.user_id != request.user.id:
        flash("You're not allowed to delete this playlist")
    else:
        playlist.delete_instance()
        flash("Playlist deleted")

    return redirect(url_for("frontend.playlist_index"))



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2017-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import unittest

from supysonic.db import Folder

from .frontendtestbase import FrontendTestBase


class FolderTestCase(FrontendTestBase):
    def test_index(self):
        self._login("bob", "B0b")
        rv = self.client.get("/folder", follow_redirects=True)
        self.assertIn("There's nothing much to see", rv.data)
        self.assertNotIn("Music folders", rv.data)
        self._logout()

        self._login("alice", "Alic3")
        rv = self.client.get("/folder")
        self.assertIn("Music folders", rv.data)

    def test_add_get(self):
        self._login("bob", "B0b")
        rv = self.client.get("/folder/add", follow_redirects=True)
        self.assertIn("There's nothing much to see", rv.data)
        self.assertNotIn("Add Folder", rv.data)
        self._logout()

        self._login("alice", "Alic3")
        rv = self.client.get("/folder/add")
        self.assertIn("Add Folder", rv.data)

    def test_add_post(self):
        self._login("alice", "Alic3")
        rv = self.client.post("/folder/add")
        self.assertIn("required", rv.data)
        rv = self.client.post("/folder/add", data={"name": "name"})
        self.assertIn("required", rv.data)
        rv = self.client.post("/folder/add", data={"path": "path"})
        self.assertIn("required", rv.data)
        rv = self.client.post("/folder/add", data={"name": "name", "path": "path"})
        self.assertIn("Add Folder", rv.data)
        rv = self.client.post(
            "/folder/add",
            data={"name": "name", "path": "tests/assets"},
            follow_redirects=True,
        )
        self.assertIn("created", rv.data)
        self.assertEqual(Folder.select().count(), 1)

    def test_delete(self):
        folder = Folder.create(name="folder", path="tests/assets", root=True)

        self._login("bob", "B0b")
        rv = self.client.get("/folder/del/" + str(folder.id), follow_redirects=True)
        self.assertIn("There's nothing much to see", rv.data)
        self.assertEqual(Folder.select().count(), 1)
        self._logout()

        self._login("alice", "Alic3")
        rv = self.client.get("/folder/del/string", follow_redirects=True)
        self.assertIn("Invalid folder id", rv.data)
        rv = self.client.get("/folder/del/1234567890", follow_redirects=True)
        self.assertIn("No such folder", rv.data)
        rv = self.client.get("/folder/del/" + str(folder.id), follow_redirects=True)
        self.assertIn("Music folders", rv.data)
        self.assertEqual(Folder.select().count(), 0)

    def test_scan(self):
        folder = Folder.create(name="folder", path="tests/assets/folder", root=True)

        self._login("alice", "Alic3")

        rv = self.client.get("/folder/scan/string", follow_redirects=True)
        self.assertIn("Invalid folder id", rv.data)
        rv = self.client.get("/folder/scan/1234567890", follow_redirects=True)
        self.assertIn("No such folder", rv.data)
        rv = self.client.get("/folder/scan/" + str(folder.id), follow_redirects=True)
        self.assertIn("start", rv.data)
        rv = self.client.get("/folder/scan", follow_redirects=True)
        self.assertIn("start", rv.data)


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2023 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import click
import time

from click.exceptions import ClickException

from .config import IniConfig
from .daemon.client import DaemonClient
from .daemon.exceptions import DaemonUnavailableError
from .db import Folder, User, init_database, release_database
from .managers.folder import FolderManager
from .managers.user import UserManager
from .scanner import Scanner


class TimedProgressDisplay:
    def __init__(self, interval=5):
        self.__stdout = click.get_text_stream("stdout")
        self.__interval = interval
        self.__last_display = 0
        self.__last_len = 0

    def __call__(self, name, scanned):
        if time.time() - self.__last_display > self.__interval:
            progress = f"Scanning '{name}': {scanned} files scanned"
            self.__stdout.write("\b" * self.__last_len)
            self.__stdout.write(progress)
            self.__stdout.flush()

            self.__last_len = len(progress)
            self.__last_display = time.time()


@click.group()
def cli():
    """Supysonic management command line interface"""
    pass


@cli.group()
def folder():
    """Folder management commands"""
    pass


@folder.command("list")
def folder_list():
    """Lists folders."""

    click.echo("Name\t\tPath\n----\t\t----")
    for f in Folder.select().where(Folder.root):
        click.echo(f"{f.name: <16}{f.path}")


@folder.command("add")
@click.argument("name")
@click.argument(
    "path",
    type=click.Path(exists=True, file_okay=False, dir_okay=True, resolve_path=True),
)
def folder_add(name, path):
    """Adds a folder.

    NAME can be anything but must be unique.
    PATH must point to an existing readable directory on the filesystem.

    If the daemon is running it will start to listen for changes in this folder but will
    not scan files already present in the folder.
    """

    try:
        FolderManager.add(name, path)
        click.echo(f"Folder '{name}' added")
    except ValueError as e:
        raise ClickException(str(e)) from e


@folder.command("delete")
@click.argument("name")
def folder_delete(name):
    """Deletes a folder.

    NAME is the name of the folder to delete.
    """

    try:
        FolderManager.delete_by_name(name)
        click.echo(f"Deleted folder '{name}'")
    except Folder.DoesNotExist as e:
        raise ClickException(f"Folder '{name}' does not exist.") from e


@folder.command("scan")
@click.argument(
    "folder",
    nargs=-1,
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force scan of already known files even if they haven't changed",
)
@click.option(
    "--background",
    "mode",
    flag_value="background",
    help="Scan the folder(s) in the background. Requires the daemon to be running.",
)
@click.option(
    "--foreground",
    "mode",
    flag_value="foreground",
    help="Scan the folder(s) in the foreground, blocking the processus while the scan is running.",
)
@click.pass_obj
def folder_scan(config, folder, force, mode):
    """Run a scan on specified folders.

    FOLDER is the name of the folder to scan. Multiple can be specified. If ommitted,
    all folders are scanned.
    """

    daemon = DaemonClient(config.DAEMON["socket"])

    # quick and dirty shorthand calls
    scan_bg = lambda: daemon.scan(folder, force)
    scan_fg = lambda: _folder_scan_foreground(config, daemon, folder, force)

    auto = not mode
    if auto:
        try:
            scan_bg()
        except DaemonUnavailableError:
            click.echo(
                "Couldn't connect to the daemon, scanning in foreground", err=True
            )
            scan_fg()
    elif mode == "background":
        try:
            scan_bg()
        except DaemonUnavailableError as e:
            raise ClickException(
                "Couldn't connect to the daemon, please use the '--foreground' option",
            ) from e
    elif mode == "foreground":
        scan_fg()


def _folder_scan_foreground(config, daemon, folders, force):
    try:
        progress = daemon.get_scanning_progress()
        if progress is not None:
            raise ClickException(
                "The daemon is currently scanning, can't start a scan now"
            )
    except DaemonUnavailableError:
        pass

    extensions = config.BASE["scanner_extensions"]
    if extensions:
        extensions = extensions.split(" ")

    def unwatch_folder(folder):
        try:
            daemon.remove_watched_folder(folder.path)
        except DaemonUnavailableError:
            pass

    def watch_folder(folder):
        try:
            daemon.add_watched_folder(folder.path)
        except DaemonUnavailableError:
            pass

    scanner = Scanner(
        force=force,
        extensions=extensions,
        follow_symlinks=config.BASE["follow_symlinks"],
        progress=TimedProgressDisplay(),
        on_folder_start=unwatch_folder,
        on_folder_end=watch_folder,
    )

    if folders:
        fstrs = folders
        folders = [
            f
            for f, in Folder.select(Folder.name)
            .where(Folder.root, Folder.name.in_(fstrs))
            .tuples()
        ]
        notfound = set(fstrs) - set(folders)
        if notfound:
            click.echo("No such folder(s): " + " ".join(notfound))
        for folder in folders:
            scanner.queue_folder(folder)
    else:
        for (folder,) in Folder.select(Folder.name).where(Folder.root).tuples():
            scanner.queue_folder(folder)

    scanner.run()
    stats = scanner.stats()

    click.echo("\nScanning done")
    click.echo(
        "Added: {0.artists} artists, {0.albums} albums, {0.tracks} tracks".format(
            stats.added
        )
    )
    click.echo(
        "Deleted: {0.artists} artists, {0.albums} albums, {0.tracks} tracks".format(
            stats.deleted
        )
    )
    if stats.errors:
        click.echo("Errors in:")
        for err in stats.errors:
            click.echo("- " + err)


@cli.group("user")
def user():
    """User management commands"""
    pass


@user.command("list")
def user_list():
    """Lists users."""

    click.echo("Name\t\tAdmin\tJukebox\tEmail")
    click.echo("----\t\t-----\t-------\t-----")
    for u in User.select():
        click.echo(
            "{: <16}{}\t{}\t{}".format(
                u.name, "*" if u.admin else "", "*" if u.jukebox else "", u.mail
            )
        )


@user.command("add")
@click.argument("name")
@click.password_option("-p", "--password", help="Specifies the user's password")
@click.option("-e", "--email", default="", help="Sets the user's email address")
def user_add(name, password, email):
    """Adds a new user.

    NAME is the name (or login) of the new user.
    """

    try:
        UserManager.add(name, password, mail=email)
    except ValueError as e:
        raise ClickException(str(e)) from e


@user.command("delete")
@click.argument("name")
def user_delete(name):
    """Deletes a user.

    NAME is the name of the user to delete.
    """

    try:
        UserManager.delete_by_name(name)
        click.echo(f"Deleted user '{name}'")
    except User.DoesNotExist as e:
        raise ClickException(f"User '{name}' does not exist.") from e


def _echo_role_change(username, name, value):
    click.echo(
        "{} '{}' {} rights".format("Granted" if value else "Revoked", username, name)
    )


@user.command("setroles")
@click.argument("name")
@click.option(
    "-A/-a", "--admin/--noadmin", default=None, help="Grant or revoke admin rights"
)
@click.option(
    "-J/-j",
    "--jukebox/--nojukebox",
    default=None,
    help="Grant or revoke jukebox rights",
)
def user_roles(name, admin, jukebox):
    """Enable/disable rights for a user.

    NAME is the login of the user to which grant or revoke rights.
    """

    try:
        user = User.get(name=name)
    except User.DoesNotExist as e:
        raise ClickException("No such user") from e

    if admin is not None:
        user.admin = admin
        _echo_role_change(name, "admin", admin)
    if jukebox is not None:
        user.jukebox = jukebox
        _echo_role_change(name, "jukebox", jukebox)
    user.save()


@user.command("changepass")
@click.argument("name")
@click.password_option("-p", "--password", help="New password")
def user_changepass(name, password):
    """Changes a user's password.

    NAME is the login of the user to which change the password.
    """

    try:
        UserManager.change_password2(name, password)
        click.echo(f"Successfully changed '{name}' password")
    except User.DoesNotExist as e:
        raise ClickException(f"User '{name}' does not exist.") from e


@user.command("rename")
@click.argument("name")
@click.argument("newname")
def user_rename(name, newname):
    """Renames a user.

    User NAME will then be known as NEWNAME.
    """

    if not name or not newname:
        raise ClickException("Missing user current name or new name")

    if name == newname:
        return

    try:
        user = User.get(name=name)
    except User.DoesNotExist as e:
        raise ClickException("No such user") from e

    try:
        User.get(name=newname)
        raise ClickException("This name is already taken")
    except User.DoesNotExist:
        pass

    user.name = newname
    user.save()
    click.echo(f"User '{name}' renamed to '{newname}'")


def main():
    config = IniConfig.from_common_locations()
    init_database(config.BASE["database_uri"])
    try:
        cli.main(obj=config)
    finally:
        release_database()


if __name__ == "__main__":
    main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2021 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import importlib
import os
import os.path
import unittest

from unittest.suite import TestSuite


def load_tests(loader, tests, pattern):
    # Skip these tests from discovery
    return tests


suite = TestSuite()
for e in os.scandir(os.path.dirname(__file__)):
    if not e.name.startswith("test") or not e.name.endswith(".py"):
        continue

    module = importlib.import_module("tests.net." + e.name[:-3])
    tests = unittest.defaultTestLoader.loadTestsFromModule(module)
    suite.addTests(tests)



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import re
import string

from flask import current_app, request
from peewee import fn

from ..db import Folder, Artist, Album, Track

from . import get_entity, get_root_folder, api_routing


@api_routing("/getMusicFolders")
def list_folders():
    return request.formatter(
        "musicFolders",
        {
            "musicFolder": [
                {"id": str(f.id), "name": f.name}
                for f in Folder.select().where(Folder.root).order_by(Folder.name)
            ]
        },
    )


def build_ignored_articles_pattern():
    articles = current_app.config["WEBAPP"]["index_ignored_prefixes"]
    if articles is None:
        return None

    articles = articles.split()
    if not articles:
        return None

    return r"^(" + r" |".join(re.escape(a) for a in articles) + r" )"


def ignored_articles_str():
    articles = current_app.config["WEBAPP"]["index_ignored_prefixes"]
    if articles is None:
        return ""

    return " ".join(articles.split())


def build_indexes(source):
    indexes = {}
    pattern = build_ignored_articles_pattern()
    for item in source:
        name = item.name
        if pattern:
            name = re.sub(pattern, "", name, flags=re.I)
        index = name[0].upper()
        if index in string.digits:
            index = "#"
        elif index not in string.ascii_letters:
            index = "?"

        if index not in indexes:
            indexes[index] = []

        indexes[index].append((item, name))

    return indexes


@api_routing("/getIndexes")
def list_indexes():
    musicFolderId = request.values.get("musicFolderId")
    ifModifiedSince = request.values.get("ifModifiedSince")
    if ifModifiedSince:
        ifModifiedSince = int(ifModifiedSince) / 1000

    if musicFolderId is None:
        folders = Folder.select().where(Folder.root)[:]
    else:
        folders = [get_root_folder(musicFolderId)]

    last_modif = max(f.last_scan for f in folders)
    if ifModifiedSince is not None and last_modif < ifModifiedSince:
        return request.formatter(
            "indexes",
            {
                "lastModified": last_modif * 1000,
                "ignoredArticles": ignored_articles_str(),
            },
        )

    # The XSD lies, we don't return artists but a directory structure
    artists = []
    children = []
    for f in folders:
        artists += f.children[:]
        children += f.tracks[:]

    indexes = build_indexes(artists)
    return request.formatter(
        "indexes",
        {
            "lastModified": last_modif * 1000,
            "ignoredArticles": ignored_articles_str(),
            "index": [
                {
                    "name": k,
                    "artist": [
                        a.as_subsonic_artist(request.user)
                        for a, _ in sorted(v, key=lambda t: t[1].lower())
                    ],
                }
                for k, v in sorted(indexes.items())
            ],
            "child": [
                c.as_subsonic_child(request.user, request.client)
                for c in sorted(children, key=lambda t: t.sort_key())
            ],
        },
    )


@api_routing("/getMusicDirectory")
def show_directory():
    res = get_entity(Folder)
    return request.formatter(
        "directory", res.as_subsonic_directory(request.user, request.client)
    )


@api_routing("/getGenres")
def list_genres():
    return request.formatter(
        "genres",
        {
            "genre": [
                {"value": genre, "songCount": sc, "albumCount": ac}
                for genre, sc, ac in Track.select(
                    Track.genre, fn.count("*"), fn.count(Track.album.distinct())
                )
                .where(Track.genre.is_null(False))
                .group_by(Track.genre)
                .tuples()
            ]
        },
    )


@api_routing("/getArtists")
def list_artists():
    mfid = request.values.get("musicFolderId")

    query = Artist.select()
    if mfid is not None:
        folder = get_root_folder(mfid)
        query = Artist.select().join(Track).where(Track.root_folder == folder)

    indexes = build_indexes(query)
    return request.formatter(
        "artists",
        {
            "ignoredArticles": ignored_articles_str(),
            "index": [
                {
                    "name": k,
                    "artist": [
                        a.as_subsonic_artist(request.user)
                        for a, _ in sorted(v, key=lambda t: t[1].lower())
                    ],
                }
                for k, v in sorted(indexes.items())
            ],
        },
    )


@api_routing("/getArtist")
def artist_info():
    res = get_entity(Artist)
    info = res.as_subsonic_artist(request.user)
    albums = set(res.albums)
    albums |= {t.album for t in res.tracks}
    info["album"] = [
        a.as_subsonic_album(request.user)
        for a in sorted(albums, key=lambda a: a.sort_key())
    ]

    return request.formatter("artist", info)


@api_routing("/getAlbum")
def album_info():
    res = get_entity(Album)
    info = res.as_subsonic_album(request.user)
    info["song"] = [
        t.as_subsonic_child(request.user, request.client)
        for t in sorted(res.tracks, key=lambda t: t.sort_key())
    ]

    return request.formatter("album", info)


@api_routing("/getSong")
def track_info():
    res = get_entity(Track)
    return request.formatter(
        "song", res.as_subsonic_child(request.user, request.client)
    )



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2017-2018 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import logging
import unittest

from supysonic.lastfm import LastFm


class LastFmTestCase(unittest.TestCase):
    """Designed only to have coverage on the most important method"""

    def test_request(self):
        logging.getLogger("supysonic.lastfm").addHandler(logging.NullHandler())
        lastfm = LastFm({"api_key": "key", "secret": "secret"}, None)

        rv = lastfm._LastFm__api_request(False, method="dummy", accents="àéèùö")
        self.assertIsInstance(rv, dict)


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2023 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

from datetime import timedelta
from flask import request
from peewee import fn, JOIN

from ..db import (
    Folder,
    Artist,
    Album,
    Track,
    StarredFolder,
    StarredArtist,
    StarredAlbum,
    StarredTrack,
    RatingFolder,
    User,
)
from ..db import now, random

from . import api_routing, get_root_folder
from .exceptions import GenericError


@api_routing("/getRandomSongs")
def rand_songs():
    size = request.values.get("size", "10")
    genre, fromYear, toYear, musicFolderId = map(
        request.values.get, ("genre", "fromYear", "toYear", "musicFolderId")
    )

    size = int(size) if size else 10
    fromYear = int(fromYear) if fromYear else None
    toYear = int(toYear) if toYear else None
    root = get_root_folder(musicFolderId)

    query = Track.select()
    if fromYear:
        query = query.where(Track.year >= fromYear)
    if toYear:
        query = query.where(Track.year <= toYear)
    if genre:
        query = query.where(Track.genre == genre)
    if root:
        query = query.where(Track.root_folder == root)

    return request.formatter(
        "randomSongs",
        {
            "song": [
                t.as_subsonic_child(request.user, request.client)
                for t in query.order_by(random()).limit(size)
            ]
        },
    )


@api_routing("/getAlbumList")
def album_list():
    ltype = request.values["type"]

    size, offset, mfid = map(request.values.get, ("size", "offset", "musicFolderId"))
    size = int(size) if size else 10
    offset = int(offset) if offset else 0
    root = get_root_folder(mfid)

    query = Folder.select().join(Track, on=Track.folder).switch().group_by(Folder.id)
    if root is not None:
        query = query.where(Track.root_folder == root)

    if ltype == "random":
        return request.formatter(
            "albumList",
            {
                "album": [
                    f.as_subsonic_child(request.user)
                    for f in query.order_by(random()).limit(size)
                ]
            },
        )
    elif ltype == "newest":
        query = query.order_by(Folder.created.desc())
    elif ltype == "highest":
        query = query.join(RatingFolder, JOIN.LEFT_OUTER).order_by(
            fn.avg(RatingFolder.rating).desc()
        )
    elif ltype == "frequent":
        query = query.order_by(fn.avg(Track.play_count).desc())
    elif ltype == "recent":
        query = query.where(Track.last_play.is_null(False)).order_by(
            fn.max(Track.last_play).desc()
        )
    elif ltype == "starred":
        query = query.join(StarredFolder).where(StarredFolder.user == request.user)
    elif ltype == "alphabeticalByName":
        query = query.order_by(Folder.name)
    elif ltype == "alphabeticalByArtist":
        parent = Folder.alias()
        query = (
            query.join(parent)
            .group_by_extend(parent.id)
            .order_by(parent.name, Folder.name)
        )
    elif ltype == "byYear":
        startyear = int(request.values["fromYear"])
        endyear = int(request.values["toYear"])
        query = query.where(
            Track.year.between(min(startyear, endyear), max(startyear, endyear))
        )
        order = fn.min(Track.year)
        if endyear < startyear:
            order = order.desc()
        query = query.order_by(order)
    elif ltype == "byGenre":
        genre = request.values["genre"]
        query = query.where(Track.genre == genre)
    else:
        raise GenericError("Unknown search type")

    return request.formatter(
        "albumList",
        {
            "album": [
                f.as_subsonic_child(request.user)
                for f in query.limit(size).offset(offset)
            ]
        },
    )


@api_routing("/getAlbumList2")
def album_list_id3():
    ltype = request.values["type"]

    size, offset, mfid = map(request.values.get, ("size", "offset", "musicFolderId"))
    size = int(size) if size else 10
    offset = int(offset) if offset else 0
    root = get_root_folder(mfid)

    query = Album.select().join(Track).group_by(Album.id)
    if root is not None:
        query = query.where(Track.root_folder == root)

    if ltype == "random":
        return request.formatter(
            "albumList2",
            {
                "album": [
                    a.as_subsonic_album(request.user)
                    for a in query.order_by(random()).limit(size)
                ]
            },
        )
    elif ltype == "newest":
        query = query.order_by(fn.min(Track.created).desc())
    elif ltype == "frequent":
        query = query.order_by(fn.avg(Track.play_count).desc())
    elif ltype == "recent":
        query = query.where(Track.last_play.is_null(False)).order_by(
            fn.max(Track.last_play).desc()
        )
    elif ltype == "starred":
        query = (
            query.switch().join(StarredAlbum).where(StarredAlbum.user == request.user)
        )
    elif ltype == "alphabeticalByName":
        query = query.order_by(Album.name)
    elif ltype == "alphabeticalByArtist":
        query = (
            query.switch()
            .join(Artist)
            .group_by_extend(Artist.id)
            .order_by(Artist.name, Album.name)
        )
    elif ltype == "byYear":
        startyear = int(request.values["fromYear"])
        endyear = int(request.values["toYear"])
        query = query.having(
            fn.min(Track.year).between(min(startyear, endyear), max(startyear, endyear))
        )
        order = fn.min(Track.year)
        if endyear < startyear:
            order = order.desc()
        query = query.order_by(order)
    elif ltype == "byGenre":
        genre = request.values["genre"]
        query = query.where(Track.genre == genre)
    else:
        raise GenericError("Unknown search type")

    return request.formatter(
        "albumList2",
        {
            "album": [
                a.as_subsonic_album(request.user)
                for a in query.limit(size).offset(offset)
            ]
        },
    )


@api_routing("/getSongsByGenre")
def songs_by_genre():
    genre = request.values["genre"]

    count, offset, mfid = map(request.values.get, ("count", "offset", "musicFolderId"))
    count = int(count) if count else 10
    offset = int(offset) if offset else 0
    root = get_root_folder(mfid)

    query = Track.select().where(Track.genre == genre)
    if root is not None:
        query = query.where(Track.root_folder == root)
    return request.formatter(
        "songsByGenre",
        {
            "song": [
                t.as_subsonic_child(request.user, request.client)
                for t in query.limit(count).offset(offset)
            ]
        },
    )


@api_routing("/getNowPlaying")
def now_playing():
    query = User.select().where(
        User.last_play.is_null(False),
        User.last_play_date > now() - timedelta(minutes=3),
    )

    return request.formatter(
        "nowPlaying",
        {
            "entry": [
                {
                    **u.last_play.as_subsonic_child(request.user, request.client),
                    "username": u.name,
                    "minutesAgo": (now() - u.last_play_date).seconds / 60,
                    "playerId": 0,
                }
                for u in query
            ]
        },
    )


@api_routing("/getStarred")
def get_starred():
    mfid = request.values.get("musicFolderId")
    root = get_root_folder(mfid)

    folders = (
        StarredFolder.select(StarredFolder.starred)
        .join(Folder)
        .join(Track, on=Track.folder)
        .where(StarredFolder.user == request.user)
        .group_by(Folder)
    )
    if root is not None:
        folders = folders.where(Folder.path.startswith(root.path))

    arq = folders.having(fn.count(Track.id) == 0)
    alq = folders.having(fn.count(Track.id) > 0)
    trq = (
        StarredTrack.select(StarredTrack.starred)
        .join(Track)
        .where(StarredTrack.user == request.user)
    )

    if root is not None:
        trq = trq.where(Track.root_folder == root)

    return request.formatter(
        "starred",
        {
            "artist": [sf.starred.as_subsonic_artist(request.user) for sf in arq],
            "album": [sf.starred.as_subsonic_child(request.user) for sf in alq],
            "song": [
                st.starred.as_subsonic_child(request.user, request.client) for st in trq
            ],
        },
    )


@api_routing("/getStarred2")
def get_starred_id3():
    mfid = request.values.get("musicFolderId")
    root = get_root_folder(mfid)

    arq = (
        StarredArtist.select(StarredArtist.starred)
        .join(Artist)
        .where(StarredArtist.user == request.user)
    )
    alq = (
        StarredAlbum.select(StarredAlbum.starred)
        .join(Album)
        .where(StarredAlbum.user == request.user)
    )
    trq = (
        StarredTrack.select(StarredTrack.starred)
        .join(Track)
        .where(StarredTrack.user == request.user)
    )

    if root is not None:
        arq = arq.join(Track).where(Track.root_folder == root)
        alq = alq.join(Track).where(Track.root_folder == root)
        trq = trq.where(Track.root_folder == root)

    return request.formatter(
        "starred2",
        {
            "artist": [sa.starred.as_subsonic_artist(request.user) for sa in arq],
            "album": [sa.starred.as_subsonic_album(request.user) for sa in alq],
            "song": [
                st.starred.as_subsonic_child(request.user, request.client) for st in trq
            ],
        },
    )



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2017-2023 Alban 'spl0k' Féron
#                    2017 Óscar García Amor
#
# Distributed under terms of the GNU AGPLv3 license.

from supysonic.db import (
    Folder,
    Album,
    Artist,
    RatingFolder,
    RatingTrack,
    StarredAlbum,
    StarredArtist,
    StarredFolder,
    StarredTrack,
    Track,
    User,
    init_database,
    release_database,
)
from supysonic.managers.folder import FolderManager

import os
import shutil
import tempfile
import unittest


class FolderManagerTestCase(unittest.TestCase):
    def setUp(self):
        # Create an empty sqlite database in memory
        init_database("sqlite:")

        # Create some temporary directories
        self.media_dir = tempfile.mkdtemp()
        self.music_dir = tempfile.mkdtemp()

    def tearDown(self):
        release_database()
        shutil.rmtree(self.media_dir)
        shutil.rmtree(self.music_dir)

    def create_folders(self):
        # Add test folders
        media = FolderManager.add("media", self.media_dir)
        music = FolderManager.add("music", self.music_dir)
        self.assertIsNotNone(media)
        self.assertIsNotNone(music)

        Folder.create(
            root=False,
            parent=music,
            name="non-root",
            path=os.path.join(self.music_dir, "subfolder"),
        )

        artist = Artist.create(name="Artist")
        album = Album.create(name="Album", artist=artist)

        Track.create(
            title="Track",
            artist=artist,
            album=album,
            disc=1,
            number=1,
            path=os.path.join(self.media_dir, "somefile"),
            folder=media,
            root_folder=media,
            duration=2,
            bitrate=320,
            last_modification=0,
        )

    def create_annotations(self):
        track = Track.select().first()
        user = User.create(name="user", password="secret", salt="ABC+", last_play=track)
        folder = Folder.get(name="media")

        RatingFolder.create(user=user, rated=folder, rating=3)
        RatingTrack.create(user=user, rated=track, rating=3)

        StarredFolder.create(user=user, starred=folder)
        StarredArtist.create(user=user, starred=track.artist_id)
        StarredAlbum.create(user=user, starred=track.album_id)
        StarredTrack.create(user=user, starred=track)

    def test_get_folder(self):
        self.create_folders()

        # Get existing folders
        for name in ["media", "music"]:
            folder = Folder.get(name=name, root=True)
            self.assertEqual(FolderManager.get(folder.id), folder)

        # Get with invalid id
        self.assertRaises(ValueError, FolderManager.get, "invalid-uuid")

        # Non-existent folder
        self.assertRaises(Folder.DoesNotExist, FolderManager.get, 1234567890)

    def test_add_folder(self):
        self.create_folders()
        self.assertEqual(Folder.select().count(), 3)

        # Create duplicate
        self.assertRaises(ValueError, FolderManager.add, "media", self.media_dir)
        self.assertEqual(Folder.select().where(Folder.name == "media").count(), 1)

        # Duplicate path
        self.assertRaises(ValueError, FolderManager.add, "new-folder", self.media_dir)
        self.assertEqual(
            Folder.select().where(Folder.path == self.media_dir).count(), 1
        )

        # Invalid path
        path = os.path.abspath("/this/not/is/valid")
        self.assertRaises(ValueError, FolderManager.add, "invalid-path", path)
        self.assertFalse(Folder.select().where(Folder.path == path).exists())

        # Subfolder of already added path
        path = os.path.join(self.media_dir, "subfolder")
        os.mkdir(path)
        self.assertRaises(ValueError, FolderManager.add, "subfolder", path)
        self.assertEqual(Folder.select().count(), 3)

        # Parent folder of an already added path
        path = os.path.join(self.media_dir, "..")
        self.assertRaises(ValueError, FolderManager.add, "parent", path)
        self.assertEqual(Folder.select().count(), 3)

    def test_delete_folder(self):
        self.create_folders()

        # Delete invalid Folder ID
        self.assertRaises(ValueError, FolderManager.delete, "invalid-uuid")
        self.assertEqual(Folder.select().count(), 3)

        # Delete non-existent folder
        self.assertRaises(Folder.DoesNotExist, FolderManager.delete, 1234567890)
        self.assertEqual(Folder.select().count(), 3)

        # Delete non-root folder
        folder = Folder.get(name="non-root")
        self.assertRaises(Folder.DoesNotExist, FolderManager.delete, folder.id)
        self.assertEqual(Folder.select().count(), 3)

        # Create some annotation to ensure foreign keys are properly handled
        self.create_annotations()

        # Delete existing folders
        for name in ["media", "music"]:
            folder = Folder.get(name=name, root=True)
            FolderManager.delete(folder.id)
            self.assertRaises(Folder.DoesNotExist, Folder.__getitem__, folder.id)

        # Even if we have only 2 root folders, non-root should never exist and be cleaned anyway
        self.assertEqual(Folder.select().count(), 0)

    def test_delete_by_name(self):
        self.create_folders()

        # Delete non-existent folder
        self.assertRaises(Folder.DoesNotExist, FolderManager.delete_by_name, "null")
        self.assertEqual(Folder.select().count(), 3)

        # Create some annotation to ensure foreign keys are properly handled
        self.create_annotations()

        # Delete existing folders
        for name in ["media", "music"]:
            FolderManager.delete_by_name(name)
            self.assertFalse(Folder.select().where(Folder.name == name).exists())


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import time

from flask import current_app, request

from ..db import Track, Album, Artist, Folder
from ..db import StarredTrack, StarredAlbum, StarredArtist, StarredFolder
from ..db import RatingTrack, RatingFolder
from ..lastfm import LastFm
from ..listenbrainz import ListenBrainz

from . import get_entity, get_entity_id, api_routing
from .exceptions import AggregateException, GenericError, MissingParameter, NotFound


def star_single(cls, starcls, eid):
    """Stars an entity

    :param cls: entity class, Folder, Artist, Album or Track
    :param starcls: matching starred class, StarredFolder, StarredArtist, StarredAlbum or StarredTrack
    :param eid: id of the entity to star
    """

    try:
        e = cls[eid]
    except cls.DoesNotExist:
        raise NotFound(f"{cls.__name__} {eid}")

    try:
        starcls[request.user, eid]
        raise GenericError(f"{cls.__name__} {eid} already starred")
    except starcls.DoesNotExist:
        pass

    starcls.create(user=request.user, starred=e)


def unstar_single(cls, starcls, eid):
    """Unstars an entity

    :param cls: entity class, Folder, Artist, Album or Track
    :param starcls: matching starred class, StarredFolder, StarredArtist, StarredAlbum or StarredTrack
    :param eid: id of the entity to unstar
    """

    starcls.delete().where(
        starcls.user == request.user, starcls.starred == eid
    ).execute()


def handle_star_request(func):
    id, albumId, artistId = map(request.values.getlist, ("id", "albumId", "artistId"))

    if not id and not albumId and not artistId:
        raise MissingParameter("id, albumId or artistId")

    errors = []
    for eid in id:
        try:
            tid = get_entity_id(Track, eid)
        except GenericError:
            tid = None
        try:
            fid = get_entity_id(Folder, eid)
        except GenericError:
            fid = None
        err = None

        if tid is None and fid is None:
            raise GenericError("Invalid ID")

        if tid is not None:
            try:
                func(Track, StarredTrack, tid)
            except Exception as e:
                err = e
        else:
            try:
                func(Folder, StarredFolder, fid)
            except Exception as e:
                err = e

        if err:
            errors.append(err)

    for alId in albumId:
        alb_id = get_entity_id(Album, alId)
        try:
            func(Album, StarredAlbum, alb_id)
        except Exception as e:
            errors.append(e)

    for arId in artistId:
        art_id = get_entity_id(Artist, arId)
        try:
            func(Artist, StarredArtist, art_id)
        except Exception as e:
            errors.append(e)

    if errors:
        raise AggregateException(errors)
    return request.formatter.empty


@api_routing("/star")
def star():
    return handle_star_request(star_single)


@api_routing("/unstar")
def unstar():
    return handle_star_request(unstar_single)


@api_routing("/setRating")
def rate():
    id = request.values["id"]
    rating = request.values["rating"]

    try:
        tid = get_entity_id(Track, id)
    except GenericError:
        tid = None
    try:
        fid = get_entity_id(Folder, id)
    except GenericError:
        fid = None
    uid = None
    rating = int(rating)

    if tid is None and fid is None:
        raise GenericError("Invalid ID")

    if not 0 <= rating <= 5:
        raise GenericError("rating must be between 0 and 5 (inclusive)")

    if rating == 0:
        if tid is not None:
            RatingTrack.delete().where(
                RatingTrack.user == request.user, RatingTrack.rated == tid
            ).execute()
        else:
            RatingFolder.delete().where(
                RatingFolder.user == request.user, RatingFolder.rated == fid
            ).execute()
    else:
        if tid is not None:
            rated = Track[tid]
            rating_cls = RatingTrack
            uid = tid
        else:
            rated = Folder[fid]
            rating_cls = RatingFolder
            uid = fid

        try:
            rating_info = rating_cls[request.user, uid]
            rating_info.rating = rating
            rating_info.save()
        except rating_cls.DoesNotExist:
            rating_cls.create(user=request.user, rated=rated, rating=rating)

    return request.formatter.empty


@api_routing("/scrobble")
def scrobble():
    res = get_entity(Track)
    t, submission = map(request.values.get, ("time", "submission"))
    t = int(t) / 1000 if t else int(time.time())

    lfm = LastFm(current_app.config["LASTFM"], request.user)
    lbz = ListenBrainz(current_app.config["LISTENBRAINZ"], request.user)

    if submission in (None, "", True, "true", "True", 1, "1"):
        lfm.scrobble(res, t)
        lbz.scrobble(res, t)
    else:
        lfm.now_playing(res)
        lbz.now_playing(res)

    return request.formatter.empty



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2017-2018 Alban 'spl0k' Féron
#                    2017 Óscar García Amor
#
# Distributed under terms of the GNU AGPLv3 license.

import base64
import flask.json
import unittest

from xml.etree import ElementTree

from ..testbase import TestBase
from ..utils import hexlify


class ApiSetupTestCase(TestBase):
    __with_api__ = True

    def setUp(self):
        super().setUp()
        self._patch_client()

    def __basic_auth_get(self, username, password):
        hashed = base64.b64encode(f"{username}:{password}".encode("utf-8"))
        headers = {"Authorization": "Basic " + hashed.decode("utf-8")}
        return self.client.get(
            "/rest/ping.view", headers=headers, query_string={"c": "tests"}
        )

    def __query_params_auth_get(self, username, password):
        return self.client.get(
            "/rest/ping.view", query_string={"c": "tests", "u": username, "p": password}
        )

    def __query_params_auth_enc_get(self, username, password):
        return self.__query_params_auth_get(username, "enc:" + hexlify(password))

    def __form_auth_post(self, username, password):
        return self.client.post(
            "/rest/ping.view", data={"c": "tests", "u": username, "p": password}
        )

    def __form_auth_enc_post(self, username, password):
        return self.__form_auth_post(username, "enc:" + hexlify(password))

    def __test_auth(self, method):
        # non-existent user
        rv = method("null", "null")
        self.assertEqual(rv.status_code, 200)
        self.assertIn('status="failed"', rv.data)
        self.assertIn('code="40"', rv.data)

        # user request with bad password
        rv = method("alice", "wrong password")
        self.assertEqual(rv.status_code, 200)
        self.assertIn('status="failed"', rv.data)
        self.assertIn('code="40"', rv.data)

        # user request
        rv = method("alice", "Alic3")
        self.assertEqual(rv.status_code, 200)
        self.assertIn('status="ok"', rv.data)

    def test_auth_basic(self):
        # No auth info
        rv = self.client.get("/rest/ping.view?c=tests")
        self.assertEqual(rv.status_code, 200)
        self.assertIn('status="failed"', rv.data)
        self.assertIn('code="10"', rv.data)

        self.__test_auth(self.__basic_auth_get)

        # Shouldn't accept 'enc:' passwords
        rv = self.__basic_auth_get("alice", "enc:" + hexlify("Alic3"))
        self.assertEqual(rv.status_code, 200)
        self.assertIn('status="failed"', rv.data)
        self.assertIn('code="40"', rv.data)

    def test_auth_query_params(self):
        self.__test_auth(self.__query_params_auth_get)
        self.__test_auth(self.__query_params_auth_enc_get)

    def test_auth_post(self):
        self.__test_auth(self.__form_auth_post)
        self.__test_auth(self.__form_auth_enc_post)

    def test_required_client(self):
        rv = self.client.get(
            "/rest/ping.view", query_string={"u": "alice", "p": "Alic3"}
        )
        self.assertIn('status="failed"', rv.data)
        self.assertIn('code="10"', rv.data)

        rv = self.client.get(
            "/rest/ping.view", query_string={"u": "alice", "p": "Alic3", "c": "tests"}
        )
        self.assertIn('status="ok"', rv.data)

    def test_format(self):
        args = {"u": "alice", "p": "Alic3", "c": "tests"}
        rv = self.client.get("/rest/getLicense.view", query_string=args)
        self.assertEqual(rv.status_code, 200)
        self.assertTrue(rv.mimetype.endswith("/xml"))  # application/xml or text/xml
        self.assertIn('status="ok"', rv.data)
        xml = ElementTree.fromstring(rv.data)
        self.assertIsNotNone(xml.find("./{http://subsonic.org/restapi}license"))

        args.update({"f": "json"})
        rv = self.client.get("/rest/getLicense.view", query_string=args)
        self.assertEqual(rv.status_code, 200)
        self.assertEqual(rv.mimetype, "application/json")
        json = flask.json.loads(rv.data)
        self.assertIn("subsonic-response", json)
        self.assertEqual(json["subsonic-response"]["status"], "ok")
        self.assertIn("license", json["subsonic-response"])

        args.update({"f": "jsonp"})
        rv = self.client.get("/rest/getLicense.view", query_string=args)
        self.assertEqual(rv.mimetype, "application/json")
        json = flask.json.loads(rv.data)
        self.assertIn("subsonic-response", json)
        self.assertEqual(json["subsonic-response"]["status"], "failed")
        self.assertEqual(json["subsonic-response"]["error"]["code"], 10)

        args.update({"callback": "dummy_cb"})
        rv = self.client.get("/rest/getLicense.view", query_string=args)
        self.assertEqual(rv.status_code, 200)
        self.assertEqual(rv.mimetype, "application/javascript")
        self.assertTrue(rv.data.startswith("dummy_cb({"))
        self.assertTrue(rv.data.endswith("})"))
        json = flask.json.loads(rv.data[9:-1])
        self.assertIn("subsonic-response", json)
        self.assertEqual(json["subsonic-response"]["status"], "ok")
        self.assertIn("license", json["subsonic-response"])

    def test_not_implemented(self):
        # Access to not implemented/unknown endpoint
        rv = self.client.get(
            "/rest/unknown", query_string={"u": "alice", "p": "Alic3", "c": "tests"}
        )
        self.assertEqual(rv.status_code, 404)
        self.assertIn('status="failed"', rv.data)
        self.assertIn('code="0"', rv.data)

        rv = self.client.post(
            "/rest/unknown", data={"u": "alice", "p": "Alic3", "c": "tests"}
        )
        self.assertEqual(rv.status_code, 404)
        self.assertIn('status="failed"', rv.data)
        self.assertIn('code="0"', rv.data)

        rv = self.client.get(
            "/rest/getVideos.view",
            query_string={"u": "alice", "p": "Alic3", "c": "tests"},
        )
        self.assertEqual(rv.status_code, 200)
        self.assertIn('status="failed"', rv.data)
        self.assertIn('code="0"', rv.data)

        rv = self.client.post(
            "/rest/getVideos.view", data={"u": "alice", "p": "Alic3", "c": "tests"}
        )
        self.assertEqual(rv.status_code, 200)
        self.assertIn('status="failed"', rv.data)
        self.assertIn('code="0"', rv.data)


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2017-2018 Alban 'spl0k' Féron
# Copyright (C) 2024 Iván Ávalos
#
# Distributed under terms of the GNU AGPLv3 license.

import logging
import unittest

from supysonic.listenbrainz import ListenBrainz

from ..frontend.frontendtestbase import FrontendTestBase


class ListenBrainzTestCase(unittest.TestCase):
    """Basic test of unauthenticated ListenBrainz API method"""

    def test_request(self):
        logging.getLogger("supysonic.listenbrainz").addHandler(logging.NullHandler())
        listenbrainz = ListenBrainz({"api_url": "https://api.listenbrainz.org/"}, None)

        user = "aavalos"
        rv = listenbrainz._ListenBrainz__api_request(
            False, "/1/search/users/?search_term={0}".format(user), token="123"
        )
        self.assertIsInstance(rv, dict)


class FrontendListenBrainzCase(FrontendTestBase):
    def test_listenbrainz_link(self):
        self._login("alice", "Alic3")
        rv = self.client.get("/user/me/listenbrainz/link", follow_redirects=True)
        self.assertIn("Missing ListenBrainz auth token", rv.data)
        rv = self.client.get(
            "/user/me/listenbrainz/link",
            query_string={"token": "abcdef"},
            follow_redirects=True,
        )
        self.assertIn("Error: ", rv.data)


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import hashlib
import logging
import requests

logger = logging.getLogger(__name__)


class LastFm:
    def __init__(self, config, user):
        if config["api_key"] is not None and config["secret"] is not None:
            self.__api_key = config["api_key"]
            self.__api_secret = config["secret"].encode("utf-8")
            self.__enabled = True
        else:
            self.__enabled = False
        self.__user = user

    def link_account(self, token):
        if not self.__enabled:
            return False, "No API key set"

        res = self.__api_request(False, method="auth.getSession", token=token)
        if not res:
            return False, "Error connecting to LastFM"
        elif "error" in res:
            return False, f"Error {res['error']}: {res['message']}"
        else:
            self.__user.lastfm_session = res["session"]["key"]
            self.__user.lastfm_status = True
            self.__user.save()
            return True, "OK"

    def unlink_account(self):
        self.__user.lastfm_session = None
        self.__user.lastfm_status = True
        self.__user.save()

    def now_playing(self, track):
        if not self.__enabled:
            return

        self.__api_request(
            True,
            method="track.updateNowPlaying",
            artist=track.album.artist.name,
            track=track.title,
            album=track.album.name,
            trackNumber=track.number,
            duration=track.duration,
        )

    def scrobble(self, track, ts):
        if not self.__enabled:
            return

        self.__api_request(
            True,
            method="track.scrobble",
            artist=track.album.artist.name,
            track=track.title,
            album=track.album.name,
            timestamp=ts,
            trackNumber=track.number,
            duration=track.duration,
        )

    def __api_request(self, write, **kwargs):
        if not self.__enabled:
            return

        if write:
            if not self.__user.lastfm_session or not self.__user.lastfm_status:
                return
            kwargs["sk"] = self.__user.lastfm_session

        kwargs["api_key"] = self.__api_key

        sig_str = b""
        for k, v in sorted(kwargs.items()):
            k = k.encode("utf-8")
            v = v.encode("utf-8") if isinstance(v, str) else str(v).encode("utf-8")
            sig_str += k + v
        sig = hashlib.md5(sig_str + self.__api_secret).hexdigest()

        kwargs["api_sig"] = sig
        kwargs["format"] = "json"

        try:
            if write:
                r = requests.post(
                    "https://ws.audioscrobbler.com/2.0/", data=kwargs, timeout=5
                )
            else:
                r = requests.get(
                    "https://ws.audioscrobbler.com/2.0/", params=kwargs, timeout=5
                )
        except requests.exceptions.RequestException as e:
            logger.warning("Error while connecting to LastFM: " + str(e))
            return None

        json = r.json()
        if "error" in json:
            if json["error"] in (9, "9"):
                self.__user.lastfm_status = False
                self.__user.save()
            logger.warning("LastFM error %i: %s", json["error"], json["message"])

        return json



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2022 Alban 'spl0k' Féron
#               2018-2019 Carey 'pR0Ps' Metcalfe
#
# Distributed under terms of the GNU AGPLv3 license.

import hashlib
import json
import logging
import mediafile
import mimetypes
import os.path
import requests
import shlex
import subprocess
import zlib

from flask import request, Response, send_file
from flask import current_app
from PIL import Image
from xml.etree import ElementTree
from zipstream import ZipStream

from ..cache import CacheMiss
from ..db import Track, Album, Artist, Folder, now
from ..covers import EXTENSIONS

from . import get_entity, get_entity_id, api_routing
from .exceptions import (
    GenericError,
    NotFound,
    ServerError,
    UnsupportedParameter,
)

logger = logging.getLogger(__name__)


def prepare_transcoding_cmdline(
    base_cmdline, res, input_format, output_format, output_bitrate
):
    if not base_cmdline:
        return None
    ret = shlex.split(base_cmdline)
    ret = [
        part.replace("%srcpath", res.path)
        .replace("%srcfmt", input_format)
        .replace("%outfmt", output_format)
        .replace("%outrate", str(output_bitrate))
        .replace("%title", res.title)
        .replace("%album", res.album.name)
        .replace("%artist", res.artist.name)
        .replace("%tracknumber", str(res.number))
        .replace("%totaltracks", str(res.album.tracks.count()))
        .replace("%discnumber", str(res.disc))
        .replace("%genre", res.genre if res.genre else "")
        .replace("%year", str(res.year) if res.year else "")
        for part in ret
    ]
    return ret


@api_routing("/stream")
def stream_media():
    res = get_entity(Track)

    if "timeOffset" in request.values:
        raise UnsupportedParameter("timeOffset")
    if "size" in request.values:
        raise UnsupportedParameter("size")

    maxBitRate, request_format, estimateContentLength = map(
        request.values.get, ("maxBitRate", "format", "estimateContentLength")
    )
    if request_format:
        request_format = request_format.lower()

    src_suffix = res.suffix()
    dst_suffix = res.suffix()
    dst_bitrate = res.bitrate
    dst_mimetype = res.mimetype

    config = current_app.config["TRANSCODING"]
    prefs = request.client

    using_default_format = False
    if request_format:
        dst_suffix = src_suffix if request_format == "raw" else request_format
    elif prefs.format:
        dst_suffix = prefs.format
    else:
        using_default_format = True
        dst_suffix = src_suffix

    if prefs.bitrate and prefs.bitrate < dst_bitrate:
        dst_bitrate = prefs.bitrate

    if maxBitRate:
        maxBitRate = int(maxBitRate)

        if dst_bitrate > maxBitRate and maxBitRate != 0:
            dst_bitrate = maxBitRate
            if using_default_format:
                dst_suffix = config.get("default_transcode_target") or dst_suffix

    # Find new mimetype if we're changing formats
    if dst_suffix != src_suffix:
        dst_mimetype = (
            mimetypes.guess_type("dummyname." + dst_suffix, False)[0]
            or "application/octet-stream"
        )

    if dst_suffix != src_suffix or dst_bitrate != res.bitrate:
        # Requires transcoding
        cache = current_app.transcode_cache
        cache_key = f"{res.id}-{dst_bitrate}.{dst_suffix}"

        try:
            response = send_file(
                cache.get(cache_key), mimetype=dst_mimetype, conditional=True
            )
        except CacheMiss:
            transcoder = config.get(f"transcoder_{src_suffix}_{dst_suffix}")
            decoder = config.get("decoder_" + src_suffix) or config.get("decoder")
            encoder = config.get("encoder_" + dst_suffix) or config.get("encoder")
            if not transcoder and (not decoder or not encoder):
                transcoder = config.get("transcoder")
                if not transcoder:
                    message = "No way to transcode from {} to {}".format(
                        src_suffix, dst_suffix
                    )
                    logger.info(message)
                    raise GenericError(message)

            transcoder, decoder, encoder = (
                prepare_transcoding_cmdline(x, res, src_suffix, dst_suffix, dst_bitrate)
                for x in (transcoder, decoder, encoder)
            )
            try:
                if transcoder:
                    dec_proc = None
                    proc = subprocess.Popen(transcoder, stdout=subprocess.PIPE)
                else:
                    dec_proc = subprocess.Popen(decoder, stdout=subprocess.PIPE)
                    proc = subprocess.Popen(
                        encoder, stdin=dec_proc.stdout, stdout=subprocess.PIPE
                    )
            except OSError:
                raise ServerError("Error while running the transcoding process")

            if estimateContentLength == "true":
                estimate = dst_bitrate * 1000 * res.duration // 8
            else:
                estimate = None

            def transcode():
                while True:
                    data = proc.stdout.read(8192)
                    if not data:
                        break
                    yield data

            def kill_processes():
                if dec_proc is not None:
                    dec_proc.kill()
                proc.kill()

            def handle_transcoding():
                try:
                    sent = 0
                    for data in transcode():
                        sent += len(data)
                        yield data
                except (Exception, SystemExit, KeyboardInterrupt):
                    # Make sure child processes are always killed
                    kill_processes()
                    raise
                except GeneratorExit:
                    # Try to transcode/send more data if we're close to the end.
                    # The calling code have to support this as yielding more data
                    # after a GeneratorExit would normally raise a RuntimeError.
                    # Hopefully this generator is only used by the cache which
                    # handles this.
                    if estimate and sent >= estimate * 0.95:
                        yield from transcode()
                    else:
                        kill_processes()
                        raise
                finally:
                    if dec_proc is not None:
                        dec_proc.stdout.close()
                        dec_proc.wait()
                    proc.stdout.close()
                    proc.wait()

            resp_content = cache.set_generated(cache_key, handle_transcoding)

            logger.info(
                "Transcoding track {0.id} for user {1.id}. Source: {2} at {0.bitrate}kbps. Dest: {3} at {4}kbps".format(
                    res, request.user, src_suffix, dst_suffix, dst_bitrate
                )
            )
            response = Response(resp_content, mimetype=dst_mimetype)
            if estimate is not None:
                response.headers.add("Content-Length", estimate)
    else:
        response = send_file(res.path, mimetype=dst_mimetype, conditional=True)

    res.play_count = res.play_count + 1
    res.last_play = now()
    res.save()

    user = request.user
    user.last_play = res
    user.last_play_date = now()
    user.save()

    return response


@api_routing("/download")
def download_media():
    id = request.values["id"]

    try:
        uid = get_entity_id(Track, id)
    except GenericError:
        uid = None
    try:
        fid = get_entity_id(Folder, id)
    except GenericError:
        fid = None

    if uid is None and fid is None:
        raise GenericError("Invalid ID")

    if uid is not None:
        try:
            rv = Track[uid]
            return send_file(rv.path, mimetype=rv.mimetype, conditional=True)
        except Track.DoesNotExist:
            try:  # Album -> stream zipped tracks
                rv = Album[uid]
            except Album.DoesNotExist as e:
                raise NotFound("Track or Album") from e
    else:
        try:  # Folder -> stream zipped tracks, non recursive
            rv = Folder[fid]
        except Folder.DoesNotExist as e:
            raise NotFound("Folder") from e

    # Stream a zip of multiple files to the client
    z = ZipStream(sized=True)
    if isinstance(rv, Folder):
        # Add the entire folder tree to the zip
        z.add_path(rv.path, recurse=True)
    else:
        # Add tracks + cover art to the zip, preventing potential naming collisions
        seen = set()
        for track in rv.tracks:
            filename = os.path.basename(track.path)
            name, ext = os.path.splitext(filename)
            index = 0
            while filename in seen:
                index += 1
                filename = f"{name} ({index})"
                if ext:
                    filename += ext

            z.add_path(track.path, filename)
            seen.add(filename)

        cover_path = _cover_from_collection(rv, extract=False)
        if cover_path:
            z.add_path(cover_path)

    if not z:
        raise GenericError("Nothing to download")

    resp = Response(z, mimetype="application/zip")
    resp.headers["Content-Disposition"] = f"attachment; filename={rv.name}.zip"
    resp.headers["Content-Length"] = len(z)
    return resp


def _cover_from_track(obj):
    """Extract and return a path to a track's cover art

    Returns None if no cover art is available.
    """
    cache = current_app.cache
    cache_key = f"{obj.id}-cover"
    try:
        return cache.get(cache_key)
    except CacheMiss:
        try:
            return cache.set(cache_key, mediafile.MediaFile(obj.path).art)
        except mediafile.UnreadableFileError:
            return None


def _cover_from_collection(obj, extract=True):
    """Get a path to cover art from a collection (Album, Folder)

    If `extract` is True, will fall back to extracting cover art from tracks
    Returns None if no cover art is available.
    """
    cover_path = None

    if isinstance(obj, Folder) and obj.cover_art:
        cover_path = os.path.join(obj.path, obj.cover_art)

    elif isinstance(obj, Album):
        track_with_folder_cover = (
            obj.tracks.join(Folder, on=Track.folder)
            .where(Folder.cover_art.is_null(False))
            .first()
        )
        if track_with_folder_cover is not None:
            cover_path = _cover_from_collection(track_with_folder_cover.folder)

        if not cover_path and extract:
            track_with_embedded = obj.tracks.where(Track.has_art).first()
            if track_with_embedded is not None:
                cover_path = _cover_from_track(track_with_embedded.id)

    if not cover_path or not os.path.isfile(cover_path):
        return None
    return cover_path


def _get_cover_path(eid):
    try:
        fid = get_entity_id(Folder, eid)
    except GenericError:
        fid = None
    try:
        uid = get_entity_id(Track, eid)
    except GenericError:
        uid = None

    if not fid and not uid:
        raise GenericError("Invalid ID")

    if fid:
        try:
            return _cover_from_collection(Folder[fid])
        except Folder.DoesNotExist:
            pass
    elif uid:
        try:
            return _cover_from_track(Track[uid])
        except Track.DoesNotExist:
            pass

        try:
            return _cover_from_collection(Album[uid])
        except Album.DoesNotExist:
            pass

    raise NotFound("Entity")


@api_routing("/getCoverArt")
def cover_art():
    cache = current_app.cache

    eid = request.values["id"]
    cover_path = _get_cover_path(eid)

    if not cover_path:
        raise NotFound("Cover art")

    size = request.values.get("size")
    if size:
        size = int(size)
    else:
        # If the cover was extracted from a track it won't have an accurate
        # extension for Flask to derive the mimetype from - derive it from the
        # contents instead.
        mimetype = None
        if os.path.splitext(cover_path)[1].lower() not in EXTENSIONS:
            with Image.open(cover_path) as im:
                mimetype = f"image/{im.format.lower()}"
        return send_file(cover_path, mimetype=mimetype)

    with Image.open(cover_path) as im:
        mimetype = f"image/{im.format.lower()}"
        if size > im.width and size > im.height:
            return send_file(cover_path, mimetype=mimetype)

        cache_key = f"{eid}-cover-{size}"
        try:
            return send_file(cache.get(cache_key), mimetype=mimetype)
        except CacheMiss:
            im.thumbnail([size, size], Image.Resampling.LANCZOS)
            with cache.set_fileobj(cache_key) as fp:
                im.save(fp, im.format)
            return send_file(cache.get(cache_key), mimetype=mimetype)


def lyrics_response_for_track(track, lyrics):
    return request.formatter(
        "lyrics",
        {"artist": track.album.artist.name, "title": track.title, "value": lyrics},
    )


@api_routing("/getLyrics")
def lyrics():
    artist = request.values["artist"]
    title = request.values["title"]

    query = (
        Track.select()
        .join(Artist)
        .where(Track.title.contains(title), Artist.name.contains(artist))
    )
    for track in query:
        # Read from track metadata
        lyrics = mediafile.MediaFile(track.path).lyrics
        if lyrics is not None:
            lyrics = lyrics.replace("\x00", "").strip()
            if lyrics:
                logger.debug("Found lyrics in file metadata: " + track.path)
                return lyrics_response_for_track(track, lyrics)

        # Look for a text file with the same name of the track
        lyrics_path = os.path.splitext(track.path)[0] + ".txt"
        if os.path.exists(lyrics_path):
            logger.debug("Found lyrics file: " + lyrics_path)

            try:
                with open(lyrics_path) as f:
                    lyrics = f.read()
            except UnicodeError:
                # Lyrics file couldn't be decoded. Rather than displaying an error, try
                # with the potential next files or return no lyrics. Log it anyway.
                logger.warning("Unsupported encoding for lyrics file " + lyrics_path)
                continue

            return lyrics_response_for_track(track, lyrics)

    if not current_app.config["WEBAPP"]["online_lyrics"]:
        return request.formatter("lyrics", {})

    # Create a stable, unique, filesystem-compatible identifier for the artist+title
    unique = hashlib.md5(
        json.dumps([x.lower() for x in (artist, title)]).encode("utf-8")
    ).hexdigest()
    cache_key = f"lyrics-{unique}"

    lyrics = {}
    try:
        lyrics = json.loads(
            zlib.decompress(current_app.cache.get_value(cache_key)).decode("utf-8")
        )
    except (CacheMiss, zlib.error, TypeError, ValueError):
        try:
            r = requests.get(
                "http://api.chartlyrics.com/apiv1.asmx/SearchLyricDirect",
                params={"artist": artist, "song": title},
                timeout=5,
            )
            root = ElementTree.fromstring(r.content)

            ns = {"cl": "http://api.chartlyrics.com/"}
            lyrics = {
                "artist": root.find("cl:LyricArtist", namespaces=ns).text,
                "title": root.find("cl:LyricSong", namespaces=ns).text,
                "value": root.find("cl:Lyric", namespaces=ns).text,
            }

            current_app.cache.set(
                cache_key, zlib.compress(json.dumps(lyrics).encode("utf-8"), 9)
            )
        except requests.exceptions.RequestException as e:  # pragma: nocover
            logger.warning("Error while requesting the ChartLyrics API: " + str(e))

    return request.formatter("lyrics", lyrics)



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2020-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import uuid

from supysonic.db import RadioStation

from .apitestbase import ApiTestBase


class RadioStationTestCase(ApiTestBase):
    def assertRadioStationCountEqual(self, count):
        self.assertEqual(RadioStation.select().count(), count)

    def assertRadioStationEquals(self, station, stream_url, name, homepage_url=None):
        self.assertEqual(station.stream_url, stream_url)
        self.assertEqual(station.name, name)
        self.assertEqual(station.homepage_url, homepage_url)

    def test_create_radio_station(self):
        # test for non-admin access
        self._make_request(
            "createInternetRadioStation",
            {"u": "bob", "p": "B0b", "username": "alice"},
            error=50,
        )

        # check params
        self._make_request("createInternetRadioStation", error=10)
        self._make_request(
            "createInternetRadioStation", {"streamUrl": "missingName"}, error=10
        )
        self._make_request(
            "createInternetRadioStation", {"name": "missing stream"}, error=10
        )

        # create w/ required fields
        stream_url = "http://example.com/radio/create"
        name = "radio station"

        self._make_request(
            "createInternetRadioStation", {"streamUrl": stream_url, "name": name}
        )

        # the correct value is 2 because _make_request uses GET then POST
        self.assertRadioStationCountEqual(2)

        for rs in RadioStation.select():
            self.assertRadioStationEquals(rs, stream_url, name)

            RadioStation.delete().execute()

        # create w/ all fields
        stream_url = "http://example.com/radio/create1"
        name = "radio station1"
        homepage_url = "http://example.com/home"

        self._make_request(
            "createInternetRadioStation",
            {"streamUrl": stream_url, "name": name, "homepageUrl": homepage_url},
        )

        # the correct value is 2 because _make_request uses GET then POST
        self.assertRadioStationCountEqual(2)

        for rs in RadioStation.select():
            self.assertRadioStationEquals(rs, stream_url, name, homepage_url)

    def test_update_radio_station(self):
        self._make_request(
            "updateInternetRadioStation",
            {"u": "bob", "p": "B0b", "username": "alice"},
            error=50,
        )

        # test data
        test = {
            "stream_url": "http://example.com/radio/update",
            "name": "Radio Update",
            "homepage_url": "http://example.com/update",
        }
        update = {
            "stream_url": test["stream_url"] + "-1",
            "name": test["name"] + "-1",
            "homepage_url": test["homepage_url"] + "-1",
        }

        # load a test record
        station = RadioStation.create(
            stream_url=test["stream_url"],
            name=test["name"],
            homepage_url=test["homepage_url"],
        )

        # check params
        self._make_request(
            "updateInternetRadioStation",
            {"id": station.id, "homepageUrl": "missing required params"},
            error=10,
        )
        self._make_request(
            "updateInternetRadioStation",
            {"id": station.id, "name": "missing streamUrl"},
            error=10,
        )
        self._make_request(
            "updateInternetRadioStation",
            {"id": station.id, "streamUrl": "missing name"},
            error=10,
        )

        # update the record w/ required fields
        self._make_request(
            "updateInternetRadioStation",
            {
                "id": station.id,
                "streamUrl": update["stream_url"],
                "name": update["name"],
            },
        )

        rs_update = RadioStation[station.id]

        self.assertRadioStationEquals(
            rs_update, update["stream_url"], update["name"], test["homepage_url"]
        )

        # update the record w/ all fields
        self._make_request(
            "updateInternetRadioStation",
            {
                "id": station.id,
                "streamUrl": update["stream_url"],
                "name": update["name"],
                "homepageUrl": update["homepage_url"],
            },
        )

        rs_update = RadioStation[station.id]

        self.assertRadioStationEquals(
            rs_update, update["stream_url"], update["name"], update["homepage_url"]
        )

    def test_delete_radio_station(self):
        # test for non-admin access
        self._make_request(
            "deleteInternetRadioStation",
            {"u": "bob", "p": "B0b", "username": "alice"},
            error=50,
        )

        # check params
        self._make_request("deleteInternetRadioStation", error=10)
        self._make_request("deleteInternetRadioStation", {"id": 1}, error=0)
        self._make_request(
            "deleteInternetRadioStation", {"id": str(uuid.uuid4())}, error=70
        )

        # delete
        station = RadioStation.create(
            stream_url="http://example.com/radio/delete",
            name="Radio Delete",
            homepage_url="http://example.com/update",
        )

        self._make_request(
            "deleteInternetRadioStation", {"id": station.id}, skip_post=True
        )

        self.assertRadioStationCountEqual(0)

    def test_get_radio_stations(self):
        test_range = 3
        for x in range(test_range):
            RadioStation.create(
                stream_url=f"http://example.com/radio-{x}",
                name=f"Radio {x}",
                homepage_url=f"http://example.com/update-{x}",
            )

        # verify happy path is clean
        self.assertRadioStationCountEqual(test_range)
        rv, child = self._make_request(
            "getInternetRadioStations", tag="internetRadioStations"
        )
        self.assertEqual(len(child), test_range)
        # This order is guaranteed to work because the api returns in order by name.
        # Test data is sequential by design.
        for x in range(test_range):
            station = child[x]
            self.assertTrue(station.get("streamUrl").endswith(f"radio-{x}"))
            self.assertTrue(station.get("name").endswith(f"Radio {x}"))
            self.assertTrue(station.get("homePageUrl").endswith(f"update-{x}"))

        # test for non-admin access
        rv, child = self._make_request(
            "getInternetRadioStations",
            {"u": "bob", "p": "B0b", "username": "alice"},
            tag="internetRadioStations",
        )

        self.assertEqual(len(child), test_range)



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2021-2023 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

from gunicorn.app.base import BaseApplication

from ._base import BaseServer


class GunicornApp(BaseApplication):
    def __init__(self, **config):
        self.__config = config

        super().__init__()

    def load_config(self):
        socket = self.__config["socket"]
        host = self.__config["host"]
        port = self.__config["port"]
        processes = self.__config["processes"]
        threads = self.__config["threads"]

        if socket is not None:
            self.cfg.set("bind", f"unix:{socket}")
        else:
            self.cfg.set("bind", f"{host}:{port}")

        if processes is not None:
            self.cfg.set("workers", processes)
        if threads is not None:
            self.cfg.set("threads", threads)


class GunicornServer(BaseServer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.__server = GunicornApp(**kwargs)
        self.__server.load = self._load_app

    def _build_kwargs(self):
        return {}

    def _run(self, **kwargs):
        return self.__server.run()


server = GunicornServer
